/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrenderstroke-iface
 * @short_description: An interface for drawing different stroked lines.
 *
 * #PgmRenderStrokeIFace is an interface to add stroking capabilities to
 * an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrenderstroke-iface.h"

static void pgm_render_stroke_iface_base_init (gpointer g_class);

static void
pgm_render_stroke_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_stroke_iface_get_type (void)
{
  static GType type = 0;

  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderStrokeIFaceClass),
	pgm_render_stroke_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };
      type = g_type_register_static (G_TYPE_INTERFACE, "PgmRenderStrokeIFace",
				     &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_stroke_set_width:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @width: The width of stroke. 0 by default.
 *
 * Sets the #width of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_set_width (PgmRenderStrokeIFace *stroke, gfloat width)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS (stroke)->set_width (stroke, width);
}

/**
 * pgm_render_stroke_get_width:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @width: The width to fill.
 *
 * Retrieves the #width of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_get_width (PgmRenderStrokeIFace *stroke, gfloat *width)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS (stroke)->get_width (stroke, width);
}

/**
 * pgm_render_stroke_set_pattern:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @pattern: The pattern of stroke. PGM_RENDER_STROKE_CONTINUOUS by
 *           default.
 *
 * Sets the #pattern of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_set_pattern (PgmRenderStrokeIFace *stroke,
			       PgmRenderStrokePattern pattern)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->set_pattern (stroke, pattern);
}

/**
 * pgm_render_stroke_get_pattern:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @pattern: The pattern to fill.
 *
 * Retrieves the #pattern of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_get_pattern (PgmRenderStrokeIFace *stroke,
			       PgmRenderStrokePattern *pattern)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->get_pattern (stroke, pattern);
}

/**
 * pgm_render_stroke_set_factor:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @factor: The factor of stroke. 3 by default.
 *
 * Sets the #factor of #stroke. If the pattern is
 * PGM_RENDER_STROKE_DASHED, the factor change the size of the dash as
 * well as the blank between each dash.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_set_factor (PgmRenderStrokeIFace *stroke, gint factor)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->set_factor (stroke, factor);
}

/**
 * pgm_render_stroke_get_factor:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @factor: The factor to fill.
 *
 * Retrieves the #factor of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_get_factor (PgmRenderStrokeIFace *stroke, gint *factor)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->get_factor (stroke, factor);
}

/**
 * pgm_render_stroke_set_color:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @r: The R color layer.
 * @g: The G color layer.
 * @b: The B color layer.
 * @a: The A color layer.
 *
 * Sets the color of #stroke to (#r, #g, #b, #a).
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_set_color (PgmRenderStrokeIFace *stroke, guchar r, guchar g,
			     guchar b, guchar a)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->set_color (stroke, r, g, b, a);
}

/**
 * pgm_render_stroke_get_color:
 * @stroke: A #PgmRenderStrokeIFace object.
 * @r: The R color layer to fill.
 * @g: The G color layer to fill.
 * @b: The B color layer to fill.
 * @a: The A color layer to fill.
 *
 * Retrieves the (#r, #g, #b, #a) color of #stroke.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_stroke_get_color (PgmRenderStrokeIFace *stroke, guchar *r,
			     guchar *g, guchar *b, guchar *a)
{
  g_return_val_if_fail (PGM_RENDER_IS_STROKE_IFACE (stroke),
			PGM_RENDER_ERROR_X);

  return PGM_RENDER_STROKE_IFACE_GET_CLASS
    (stroke)->get_color (stroke, r, g, b, a);
}
