/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render base library
 *
 * Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 */

/**
 * SECTION:pgmrendertext-iface
 * @short_description: An interface for drawing text.
 *
 * #PgmRenderTextIFace is an interface to add text drawing capabilities
 * to an object.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "pgmrendertext-iface.h"

static void pgm_render_text_iface_base_init (gpointer g_class);

static void
pgm_render_text_iface_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (! initialized)
    {
      initialized = TRUE;
    }
}

GType
pgm_render_text_iface_get_type (void)
{
  static GType type = 0;

  if (! type)
    {
      static const GTypeInfo info = {
	sizeof (PgmRenderTextIFaceClass),
	pgm_render_text_iface_base_init, /* base init */
	NULL, /* base finalize */
	NULL, /* class_init */
	NULL, /* class finalize */
	NULL, /* class data */
	0,
	0,    /* n_preallocs */
	NULL, /* instance init */
      };
      type = g_type_register_static (G_TYPE_INTERFACE, "PgmRenderTextIFace",
				     &info, 0);
      g_type_interface_add_prerequisite (type, G_TYPE_OBJECT);
    }

  return type;
}

/**
 * pgm_render_text_set_label:
 * @text: A #PgmRenderTextIFace object.
 * @label: The label of the text.
 *
 * Sets the #label of #text.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_text_set_label (PgmRenderTextIFace *text, const gchar *label)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS (text)->set_label (text, label);
}

/**
 * pgm_render_text_get_label:
 * @text: A #PgmRenderTextIFace object.
 * @label: The label of the text to fill.
 *
 * Retrieves the #label of #text.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_text_get_label (PgmRenderTextIFace *text, gchar *label)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS (text)->get_label (text, label);
}

/**
 * pgm_render_text_set_font:
 * @text: A #PgmRenderTextIFace object.
 * @name: The name of the font.
 * @weight: The weight type of the font.
 * @slant: The slant type of the font.
 * @size: The size of the font in [10, 100].
 *
 * Set the #font of #text.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_text_set_font (PgmRenderTextIFace *text, const gchar *name,
			  PgmRenderTextWeight weight, PgmRenderTextSlant slant,
			  gint size)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS
    (text)->set_font (text, name, weight, slant, size);
}

/**
 * pgm_render_text_get_font:
 * @text: A #PgmRenderTextIFace object.
 * @weight: The weight type of the font to fill.
 * @slant: The slant type of the font to fill.
 * @size: The size of the font to fill.
 *
 * Retrieves the #font of #text.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_text_get_font (PgmRenderTextIFace *text, PgmRenderTextWeight *weight,
			  PgmRenderTextSlant *slant, gint *size)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS
    (text)->get_font (text, weight, slant, size);
}

/**
 * pgm_render_text_set_color:
 * @text: A #PgmRenderTextIFace object.
 * @r: The R color layer.
 * @g: The G color layer.
 * @b: The B color layer.
 * @a: The A color layer.
 *
 * Sets the color of #text to (#r, #g, #b, #a).
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_text_set_color (PgmRenderTextIFace *text, guchar r, guchar g,
			   guchar b, guchar a)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS
    (text)->set_color (text, r, g, b, a);
}

/**
 * pgm_render_text_get_color:
 * @text: A #PgmRenderTextIFace object.
 * @r: The R color layer to fill.
 * @g: The G color layer to fill.
 * @b: The B color layer to fill.
 * @a: The A color layer to fill.
 *
 * Retrieves the color of #text.
 *
 * Returns: An error status. 
 */
PgmRenderError
pgm_render_text_get_color (PgmRenderTextIFace *text, guchar *r, guchar *g,
			   guchar *b, guchar *a)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS
    (text)->get_color (text, r, g, b, a);
}

/**
 * pgm_render_text_get_size:
 * @text: A #PgmRenderTextIFace object.
 * @text_width: The real width of the font to fill.
 * @text_height: The real height of the font to fill.
 *
 * Retrieves the real size (#text_width, #text_height) of the
 * generated #text.
 *
 * Returns: An error status.
 */
PgmRenderError
pgm_render_text_get_size (PgmRenderTextIFace *text, gint *text_width,
			  gint *text_height)
{
  g_return_val_if_fail (PGM_RENDER_IS_TEXT_IFACE (text), PGM_RENDER_ERROR_X);

  return PGM_RENDER_TEXT_IFACE_GET_CLASS
    (text)->get_size (text, text_width, text_height);
}
