/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment Render OpenGL 1.2 plugin
 *
 * Copyright © 2006, 2007 Fluendo Embedded S.L. <www.fluendo.com>
 * All rights reserved.
 *
 * This file is available under one of two license agreements.
 *
 * This file is licensed under the GPL version 2. See "LICENSE.GPL" in
 * the root of this distribution including a special exception to use
 * Pigment with Fluendo's plugins.
 *
 * The GPL part of Pigment is also available under a commercial licensing
 * agreement from Fluendo. See "LICENSE.Pigment" in the root directory
 * of this distribution package for details on that license.
 *
 * Author: Loïc Molinari <loic@fluendo.com>
 */

#include "pgmrendergl1surface.h"
#include "pgmrendergl1window.h"
#include <math.h>

/* Supported RGB caps on all the GL implementations */
#define PGM_RENDER_GL1_CAPS_RGB GST_VIDEO_CAPS_RGB ";"  \
                                GST_VIDEO_CAPS_BGR ";"  \
                                GST_VIDEO_CAPS_RGBA ";" \
                                GST_VIDEO_CAPS_BGRA
/* Supported YUV caps on the programmable GL implementations */
#define PGM_RENDER_GL1_CAPS_YUV GST_VIDEO_CAPS_YUV ("I420") ";" \
                                GST_VIDEO_CAPS_YUV ("IYUV") ";" \
                                GST_VIDEO_CAPS_YUV ("YV12") ";" \
                                GST_VIDEO_CAPS_YUV ("YUYV") ";" \
                                GST_VIDEO_CAPS_YUV ("YUY2") ";" \
                                GST_VIDEO_CAPS_YUV ("V422") ";" \
                                GST_VIDEO_CAPS_YUV ("YUNV") ";" \
                                GST_VIDEO_CAPS_YUV ("UYVY") ";" \
                                GST_VIDEO_CAPS_YUV ("Y422") ";" \
                                GST_VIDEO_CAPS_YUV ("UYNV")

/* Default starting max 2d texture size. This value is used by all the surfaces
 * created before the GL context setup. The OpenGL specification tells that the
 * minimum texture 2d size must be 64. 256 seems to be the minimum supported by
 * the implementations. This assertion still need to be validated though. */
#define START_UP_MAX_2D_TEXTURE_SIZE 256

static void pgm_render_gl1_surface_class_init (PgmRenderGl1SurfaceClass *klass);
static void pgm_render_gl1_surface_init (PgmRenderGl1Surface *obj);
static void pgm_render_gl1_surface_finalize (GObject *obj);

static void layout_iface_init (PgmRenderLayoutIFaceClass *iface);
static void media_iface_init (PgmRenderMediaIFaceClass *iface);
static void shape_iface_init (PgmRenderShapeIFaceClass *iface);
static void stroke_iface_init (PgmRenderStrokeIFaceClass *iface);
static void pipe3d_iface_init (PgmRenderPipe3dIFaceClass *iface);

static PgmRenderError gl1_surface_imedia_del_image (PgmRenderMediaIFace *obj);
static PgmRenderError pgm_render_gl1_surface_draw (PgmRenderDrawable *drawable);
static void set_pixel_aspect_ratio (PgmRenderGl1Surface *surface);

static PgmRenderDrawableClass *parent_class = NULL;

G_DEFINE_TYPE_WITH_CODE (PgmRenderGl1Surface, pgm_render_gl1_surface,
			 PGM_RENDER_TYPE_DRAWABLE,
			 {
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_LAYOUT_IFACE, layout_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_MEDIA_IFACE, media_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_SHAPE_IFACE, shape_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_STROKE_IFACE, stroke_iface_init);
			   G_IMPLEMENT_INTERFACE
			     (PGM_RENDER_TYPE_PIPE3D_IFACE, pipe3d_iface_init);
			 });

static void
pgm_render_gl1_surface_class_init (PgmRenderGl1SurfaceClass *klass)
{
  GObjectClass *gobject_class;
  PgmRenderDrawableClass *drawable_class;

  gobject_class = (GObjectClass*) klass;
  drawable_class = (PgmRenderDrawableClass*) klass;

  /* The surface supports at least the RGB color spaces, we append the
   * other YUV color spaces after having check if the OpenGL implementation
   * supports GLSL or the fragment_program extension. */
  drawable_class->caps = gst_caps_from_string (PGM_RENDER_GL1_CAPS_RGB);
  drawable_class->draw = pgm_render_gl1_surface_draw;

  gobject_class->finalize = pgm_render_gl1_surface_finalize;

  klass->media_format_lut[PGM_RENDER_MEDIA_RGB] = PGM_RENDER_GL_RGB;
  klass->media_format_lut[PGM_RENDER_MEDIA_BGR] = PGM_RENDER_GL_BGR;
  klass->media_format_lut[PGM_RENDER_MEDIA_RGBA] = PGM_RENDER_GL_RGBA;
  klass->media_format_lut[PGM_RENDER_MEDIA_BGRA] = PGM_RENDER_GL_BGRA;
  klass->media_format_lut[PGM_RENDER_MEDIA_I420] = PGM_RENDER_GL_LUMINANCE;
  klass->media_format_lut[PGM_RENDER_MEDIA_YV12] = PGM_RENDER_GL_LUMINANCE;
  klass->media_format_lut[PGM_RENDER_MEDIA_UYVY] = PGM_RENDER_GL_RGBA;
  klass->media_format_lut[PGM_RENDER_MEDIA_YUYV] = PGM_RENDER_GL_RGBA;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_RGB] = 3.0f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_BGR] = 3.0f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_RGBA] = 4.0f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_BGRA] = 4.0f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_I420] = 1.5f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_YV12] = 1.5f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_UYVY] = 2.0f;
  klass->media_format_lut_size[PGM_RENDER_MEDIA_YUYV] = 2.0f;
  klass->surface_with_media_data = NULL;
  klass->surface_with_new_media_interp = NULL;
  klass->texture_trash = NULL;
}

static void
init_image_data (PgmRenderGl1Surface *surface)
{
  surface->media_width = 0;
  surface->media_height = 0;
  surface->media_ratio = 1.0f;
  surface->media_pot_width = 0;
  surface->media_pot_height = 0;
  surface->media_texture = NULL;
  surface->media_tile_vertex = NULL;
  surface->media_tile_tex_coord = NULL;
  surface->media_nb_width_tiles = 0;
  surface->media_nb_height_tiles = 0;
  surface->media_nb_tiles = 0;
  surface->media_width_texel_size = 0.0f;
  surface->media_height_texel_size = 0.0f;
  surface->media_last_tile_s = 1.0f;
  surface->media_last_tile_t = 1.0f;
  surface->media_master = surface;
  surface->media_slave = NULL;
  surface->media_loaded = FALSE;
  surface->media_gst_loaded = FALSE;
  surface->media_interp = PGM_RENDER_MEDIA_BILINEAR;
  surface->media_data = NULL;
  surface->media_gst_buffer = NULL;
  surface->max_texture_size = START_UP_MAX_2D_TEXTURE_SIZE;
}

static void
pgm_render_gl1_surface_init (PgmRenderGl1Surface *obj)
{
  guint i;

  for (i = 0; i < 4; i++)
    {
      obj->vertex[i].x = 0.0f;
      obj->vertex[i].y = 0.0f;
      obj->vertex[i].z = 0.0f;
      obj->vertex[i].w = 1.0f;
      obj->color[i].r = 1.0f;
      obj->color[i].g = 1.0f;
      obj->color[i].b = 1.0f;
      obj->color[i].a = 1.0f;
    }
  obj->layout_width = 0.0f;
  obj->layout_height = 0.0f;
  obj->layout_x_position = 0.0f;
  obj->layout_y_position = 0.0f;
  obj->layout_ratio = 0.0f;
  obj->layout_x_align = 0.5f;
  obj->layout_y_align = 0.5f;
  obj->layout_style = PGM_RENDER_LAYOUT_SCALED;
  obj->layout_custom_ratio = 1.0f;

  /* Shape interface initialization */
  obj->arc_width = 0.0f;
  obj->arc_height = 0.0f;
  obj->use_arc = FALSE;

  /* Stroke interface initialization */
  obj->stroke_width = 0.0f;
  obj->stroke_color.r = 1.0f;
  obj->stroke_color.g = 1.0f;
  obj->stroke_color.b = 1.0f;
  obj->stroke_color.a = 1.0f;
  obj->stroke_pattern = PGM_RENDER_STROKE_CONTINUOUS;
  obj->stroke_factor = 1;

  /* Pipe3d interface initialization */
  obj->translation[0] = 0.0f;
  obj->translation[1] = 0.0f;
  obj->translation[2] = 0.0f;
  obj->rotation[0] = 0.0f;
  obj->rotation[1] = 0.0f;
  obj->rotation[2] = 0.0f;

  obj->sx = 1.0f;
  obj->sy = 1.0f;

  /* Media interface initialization */
  init_image_data (obj);
}

static void
pgm_render_gl1_surface_finalize (GObject *obj)
{
  G_OBJECT_CLASS (parent_class)->finalize (obj);
}

PgmRenderGl1Surface *
pgm_render_gl1_surface_new (void)
{
  PgmRenderGl1Surface *gl1_surface;

  gl1_surface = PGM_RENDER_GL1_SURFACE
    (g_object_new (PGM_RENDER_TYPE_GL1_SURFACE, NULL));

  return gl1_surface;
}

static void
set_pixel_aspect_ratio (PgmRenderGl1Surface *surface)
{
  PgmRenderBaseWindow *window;
  gfloat x_position, y_position = 0.0f;
  gfloat new_media_width, new_media_height, tmp;
  gfloat width_begin_position = 0.0f, height_begin_position = 0.0f;
  gfloat x_end = 0.0f, y_end = 0.0f;
  gfloat x_increment = 0.0f, y_increment = 0.0f;
  gint last_width_tile, last_height_tile;
  gint offset;
  gint i, j;
/*   gfloat width_aspect_factor, height_aspect_factor; */

  if (!surface->media_nb_tiles)
    return;

  window = PGM_RENDER_DRAWABLE (surface)->window;
  if (window)
    {
      /* Retrieve the aspect correction factors */
/*       width_aspect_factor = window->width_aspect_factor; */
/*       height_aspect_factor = window->height_aspect_factor; */
      surface->media_ratio = ((gfloat) surface->media_width * surface->sx/*  * width_aspect_factor */)
        / ((gfloat) surface->media_height * surface->sy/*  * height_aspect_factor */);
    }

  /* FIXME: Handle the CENTERED style */
  if (surface->layout_style == PGM_RENDER_LAYOUT_SCALED)
    {
      /* Compute the new media vertices position and size */
      if (surface->media_ratio < surface->layout_ratio)
        {
          /* Width */
          new_media_width = (gfloat) surface->layout_height * surface->sy
            * surface->media_ratio;
          tmp = ((gfloat) surface->layout_width * surface->sx - new_media_width)
            * surface->layout_x_align;
          width_begin_position = surface->vertex[0].x + tmp;
          /* Height */
          new_media_height = (gfloat) surface->layout_height * surface->sy;
          height_begin_position = surface->vertex[0].y;
        }
      else
        {
          /* Width */
          new_media_width = (gfloat) surface->layout_width * surface->sx;
          width_begin_position = surface->vertex[0].x;
          /* Height */
          new_media_height = (gfloat) surface->layout_width * surface->sx
            / surface->media_ratio;
          tmp = ((gfloat) surface->layout_height * surface->sy - new_media_height)
            * surface->layout_y_align;
          height_begin_position = surface->vertex[0].y + tmp;
        }

      x_increment = ((gfloat) surface->max_texture_size
                     / (gfloat) surface->media_width) * new_media_width;
      y_increment = ((gfloat) surface->max_texture_size
                     / (gfloat) surface->media_height) * new_media_height;

      x_end = width_begin_position + new_media_width;
      y_end = height_begin_position + new_media_height;
    }
  else if (surface->layout_style == PGM_RENDER_LAYOUT_FILLED)
    {
      width_begin_position = surface->vertex[0].x;
      height_begin_position = surface->vertex[0].y;

      x_increment = ((gfloat) surface->max_texture_size
        / (gfloat) surface->media_width) * (gfloat) surface->layout_width
        * surface->sx;
      y_increment = ((gfloat) surface->max_texture_size
        / (gfloat) surface->media_height) * (gfloat) surface->layout_height
        * surface->sy;

      x_end = surface->vertex[2].x;
      y_end = surface->vertex[1].y;
    }

  last_width_tile = surface->media_nb_width_tiles - 1;
  last_height_tile = surface->media_nb_height_tiles - 1;

  /* Set the position of the max sized tiles */
  x_position = width_begin_position;
  for (i = 0; i < last_width_tile; i++)
    {
      y_position = height_begin_position;
      for (j = 0; j < last_height_tile; j++)
        {
          offset = i * surface->media_nb_height_tiles * 12 + j * 12;
          *(surface->media_tile_vertex + offset + 0) = x_position;
          *(surface->media_tile_vertex + offset + 1) = y_position;
          *(surface->media_tile_vertex + offset + 2) = 0.0f;
          *(surface->media_tile_vertex + offset + 3) = x_position;
          *(surface->media_tile_vertex + offset + 4) = y_position + y_increment;
          *(surface->media_tile_vertex + offset + 5) = 0.0f;
          *(surface->media_tile_vertex + offset + 6) = x_position + x_increment;
          *(surface->media_tile_vertex + offset + 7) = y_position + y_increment;
          *(surface->media_tile_vertex + offset + 8) = 0.0f;
          *(surface->media_tile_vertex + offset + 9) = x_position + x_increment;
          *(surface->media_tile_vertex + offset + 10) = y_position;
          *(surface->media_tile_vertex + offset + 11) = 0.0f;
          y_position += y_increment;
        }
      x_position += x_increment;
    }

  /* Set the position of the last width tiles */
  y_position = height_begin_position;
  for (i = 0; i < last_height_tile; i++)
    {
      offset = last_width_tile * surface->media_nb_height_tiles * 12 + i * 12;
      *(surface->media_tile_vertex + offset + 0) = x_position;
      *(surface->media_tile_vertex + offset + 1) = y_position;
      *(surface->media_tile_vertex + offset + 2) = 0.0f;
      *(surface->media_tile_vertex + offset + 3) = x_position;
      *(surface->media_tile_vertex + offset + 4) = y_position + y_increment;
      *(surface->media_tile_vertex + offset + 5) = 0.0f;
      *(surface->media_tile_vertex + offset + 6) = x_end;
      *(surface->media_tile_vertex + offset + 7) = y_position + y_increment;
      *(surface->media_tile_vertex + offset + 8) = 0.0f;
      *(surface->media_tile_vertex + offset + 9) = x_end;
      *(surface->media_tile_vertex + offset + 10) = y_position;
      *(surface->media_tile_vertex + offset + 11) = 0.0f;
      y_position += y_increment;
    }

  /* Set the position of the last height tiles */
  x_position = width_begin_position;
  for (i = 0; i < last_width_tile; i++)
    {
      offset = last_height_tile * 12 + i * surface->media_nb_height_tiles * 12;
      *(surface->media_tile_vertex + offset + 0) = x_position;
      *(surface->media_tile_vertex + offset + 1) = y_position;
      *(surface->media_tile_vertex + offset + 2) = 0.0f;
      *(surface->media_tile_vertex + offset + 3) = x_position;
      *(surface->media_tile_vertex + offset + 4) = y_end;
      *(surface->media_tile_vertex + offset + 5) = 0.0f;
      *(surface->media_tile_vertex + offset + 6) = x_position + x_increment;
      *(surface->media_tile_vertex + offset + 7) = y_end;
      *(surface->media_tile_vertex + offset + 8) = 0.0f;
      *(surface->media_tile_vertex + offset + 9) = x_position + x_increment;
      *(surface->media_tile_vertex + offset + 10) = y_position;
      *(surface->media_tile_vertex + offset + 11) = 0.0f;
      x_position += x_increment;
    }

  /* Set the position of the last tile */
  offset = (surface->media_nb_tiles - 1) * 12;
  *(surface->media_tile_vertex + offset + 0) = x_position;
  *(surface->media_tile_vertex + offset + 1) = y_position;
  *(surface->media_tile_vertex + offset + 2) = 0.0f;
  *(surface->media_tile_vertex + offset + 3) = x_position;
  *(surface->media_tile_vertex + offset + 4) = y_end;
  *(surface->media_tile_vertex + offset + 5) = 0.0f;
  *(surface->media_tile_vertex + offset + 6) = x_end;
  *(surface->media_tile_vertex + offset + 7) = y_end;
  *(surface->media_tile_vertex + offset + 8) = 0.0f;
  *(surface->media_tile_vertex + offset + 9) = x_end;
  *(surface->media_tile_vertex + offset + 10) = y_position;
  *(surface->media_tile_vertex + offset + 11) = 0.0f;
}

/* Layout interface implementation */

static PgmRenderError
gl1_surface_ilayout_set_size (PgmRenderLayoutIFace *obj, gint width,
			      gint height)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->layout_width = width;
  gl1_surface->layout_height = height;
  gl1_surface->layout_ratio = (gfloat) gl1_surface->layout_width /
    (gfloat) gl1_surface->layout_height;

  gl1_surface->vertex[2].x = gl1_surface->vertex[3].x
    = gl1_surface->vertex[0].x + (gfloat) width * gl1_surface->sx;
  gl1_surface->vertex[1].y = gl1_surface->vertex[2].y
    = gl1_surface->vertex[0].y + (gfloat) height * gl1_surface->sy;

  if (gl1_surface->media_loaded || gl1_surface->media_gst_loaded)
    set_pixel_aspect_ratio (gl1_surface);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_get_size (PgmRenderLayoutIFace *obj, gint *width,
			      gint *height)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  *width = gl1_surface->layout_width;
  *height = gl1_surface->layout_height;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_set_position (PgmRenderLayoutIFace *obj, gint x, gint y)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  gint scaled_x, scaled_y;
  gint x_offset, y_offset;
  gint i, j;
  gint tile_width;

  surface->layout_x_position = x;
  surface->layout_y_position = y;
  scaled_x = (gint) (x * surface->sx);
  scaled_y = (gint) (y * surface->sy);

  /* Add an offset to the layout vertices */
  surface->vertex[0].x = surface->vertex[1].x = scaled_x;
  surface->vertex[2].x = surface->vertex[3].x =
    scaled_x + surface->layout_width * surface->sx;
  surface->vertex[0].y = surface->vertex[3].y = scaled_y;
  surface->vertex[1].y = surface->vertex[2].y =
    scaled_y + surface->layout_height * surface->sy;

  /* Add an offset to the media vertices */
  tile_width = surface->media_nb_width_tiles * 12;
  x_offset = x * surface->sx - surface->vertex[0].x;
  y_offset = y * surface->sy - surface->vertex[0].y;
  for (i = 0; i < surface->media_nb_width_tiles; i++)
    {
      for (j = 0; j < surface->media_nb_height_tiles; j++)
        {
          gint offset = i * 12 + j * tile_width;
          *(surface->media_tile_vertex + offset + 0) += x_offset;
          *(surface->media_tile_vertex + offset + 1) += y_offset;
          *(surface->media_tile_vertex + offset + 3) += x_offset;
          *(surface->media_tile_vertex + offset + 4) += y_offset;
          *(surface->media_tile_vertex + offset + 6) += x_offset;
          *(surface->media_tile_vertex + offset + 7) += y_offset;
          *(surface->media_tile_vertex + offset + 9) += x_offset;
          *(surface->media_tile_vertex + offset + 10) += y_offset;
        }
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_get_position (PgmRenderLayoutIFace *obj, gint *x, gint *y)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);

  *x = gl1_surface->vertex[0].x;
  *y = gl1_surface->vertex[0].y;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_set_alignment (PgmRenderLayoutIFace *obj, gfloat x,
				   gfloat y)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->layout_x_align = CLAMP (x, 0.0f, 1.0f);
  gl1_surface->layout_y_align = CLAMP (y, 0.0f, 1.0f);

  set_pixel_aspect_ratio (gl1_surface);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_get_alignment (PgmRenderLayoutIFace *obj, gfloat *x,
				   gfloat *y)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);

  *x = gl1_surface->layout_x_align;
  *y = gl1_surface->layout_y_align;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_set_aspect_style (PgmRenderLayoutIFace *obj,
				      PgmRenderLayoutAspectStyle style)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);

  surface->layout_style = style;
  set_pixel_aspect_ratio (surface);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_get_aspect_style (PgmRenderLayoutIFace *obj,
				      PgmRenderLayoutAspectStyle *style)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (style, PGM_RENDER_ERROR_X);

  *style = surface->layout_style;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_set_aspect_ratio (PgmRenderLayoutIFace *obj, gfloat ratio)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);

  surface->layout_custom_ratio = ratio;
  /* update ratio */

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ilayout_get_aspect_ratio (PgmRenderLayoutIFace *obj,
				      gfloat *ratio)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (ratio, PGM_RENDER_ERROR_X);

  *ratio = gl1_surface->layout_custom_ratio;

  return PGM_RENDER_ERROR_X;
}

static void
layout_iface_init (PgmRenderLayoutIFaceClass *iface)
{
  iface->set_size = gl1_surface_ilayout_set_size;
  iface->get_size = gl1_surface_ilayout_get_size;
  iface->set_position = gl1_surface_ilayout_set_position;
  iface->get_position = gl1_surface_ilayout_get_position;
  iface->set_alignment = gl1_surface_ilayout_set_alignment;
  iface->get_alignment = gl1_surface_ilayout_get_alignment;
  iface->set_aspect_style = gl1_surface_ilayout_set_aspect_style;
  iface->get_aspect_style = gl1_surface_ilayout_get_aspect_style;
  iface->set_aspect_ratio = gl1_surface_ilayout_set_aspect_ratio;
  iface->get_aspect_ratio = gl1_surface_ilayout_get_aspect_ratio;
}

/* Media interface implementation */

static PgmRenderError
gl1_surface_imedia_set_color (PgmRenderMediaIFace *obj, guchar r, guchar g,
			      guchar b, guchar a)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);
  Color converted_color;
  gint i;

  converted_color.r = (gfloat) r * INV_255;
  converted_color.g = (gfloat) g * INV_255;
  converted_color.b = (gfloat) b * INV_255;
  converted_color.a = (gfloat) a * INV_255;

  for (i = 0; i < 4; i++)
    {
      gl1_surface->color[i].r = converted_color.r;
      gl1_surface->color[i].g = converted_color.g;
      gl1_surface->color[i].b = converted_color.b;
      gl1_surface->color[i].a = converted_color.a;
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_get_color (PgmRenderMediaIFace *obj, guchar *r, guchar *g,
			      guchar *b, guchar *a)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (r, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (g, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (b, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (a, PGM_RENDER_ERROR_X);

  *r = (guchar) (gl1_surface->color[0].r * 255);
  *g = (guchar) (gl1_surface->color[0].g * 255);
  *b = (guchar) (gl1_surface->color[0].b * 255);
  *a = (guchar) (gl1_surface->color[0].a * 255);

  return PGM_RENDER_ERROR_X;
}

static gint
get_upper_power_of_two (gint number)
{
  gint power_of_two = 1;

  while (power_of_two < number)
    power_of_two <<= 1;

  return power_of_two;
}

static void
set_texture_coordinates (PgmRenderGl1Surface *surface, gfloat width_offset,
                         gfloat height_offset)
{
  gint offset, tmp;
  gint i, j;
  gfloat coord;

  /* Is there a media actually set ? */
  if (!surface->media_tile_tex_coord)
    return;

  /* Texture coordinates initialisation of the tile(s) */
  for (i = 0; i < surface->media_nb_width_tiles; i++)
    {
      for (j = 0; j < surface->media_nb_height_tiles; j++)
        {
          offset = i * surface->media_nb_height_tiles * 8;
          *(surface->media_tile_tex_coord + offset + j*8 + 0) = 0.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 1) = 0.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 2) = 0.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 3) = 1.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 4) = 1.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 5) = 1.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 6) = 1.0f;
          *(surface->media_tile_tex_coord + offset + j*8 + 7) = 0.0f;
        }
    }
  /* Texture coordinates initialisation of the first tile(s) in the width */
  for (i = 0; i < surface->media_nb_height_tiles; i++)
    {
      *(surface->media_tile_tex_coord + i*8 + 0) = width_offset;
      *(surface->media_tile_tex_coord + i*8 + 2) = width_offset;
    }
  /* Texture coordinates initialisation of the last tile(s) in the width */
  offset = (surface->media_nb_width_tiles - 1)
    * surface->media_nb_height_tiles * 8;
  for (i = 0; i < surface->media_nb_height_tiles; i++)
    {
      coord = surface->media_last_tile_s - width_offset;
      *(surface->media_tile_tex_coord + offset + i*8 + 4) = coord;
      *(surface->media_tile_tex_coord + offset + i*8 + 6) = coord;
    }
  /* Texture coordinates initialisation of the first tile(s) in the height */
  tmp = surface->media_nb_height_tiles * 8;
  for (i = 0; i < surface->media_nb_width_tiles; i++)
    {
      *(surface->media_tile_tex_coord + i*tmp + 1) = height_offset;
      *(surface->media_tile_tex_coord + i*tmp + 7) = height_offset;
    }
  /* Texture coordinates initialisation of the last tile(s) in the height */
  offset = (surface->media_nb_height_tiles - 1) * 8;
  for (i = 0; i < surface->media_nb_width_tiles; i++)
    {
      coord = surface->media_last_tile_t - height_offset;
      *(surface->media_tile_tex_coord + offset + i*tmp + 3) = coord;
      *(surface->media_tile_tex_coord + offset + i*tmp + 5) = coord;
    }
}

static void
generate_geometry_tiling (PgmRenderGl1Surface *surface,
                          PgmRenderMediaPixelFormat format, gint width,
                          gint height, gint stride)
{
  PgmRenderGl1Window *window;
  gint width_modulo, height_modulo;
  gint width_upper_pot, height_upper_pot;
  gfloat size;

  /* Retrieve the max texture size */
  window = PGM_RENDER_GL1_WINDOW (PGM_RENDER_DRAWABLE (surface)->window);
  if (window)
    surface->max_texture_size = window->backend->max_texture_2d_size;    

  /* The number of tile on the width, and the last s texture coordinate value
   * on the width. */
  surface->media_nb_width_tiles = width / surface->max_texture_size;
  width_upper_pot = get_upper_power_of_two (width);
  size = MIN (surface->max_texture_size, width_upper_pot);
  surface->media_width_texel_size = 1.0f / size;
  width_modulo = width % surface->max_texture_size;
  if (width_modulo)
    {
      surface->media_nb_width_tiles += 1;
      surface->media_last_tile_s = width_modulo / size;
    }

  /* The number of tile on the height, and the last t texture coordinate value
   * on the height. */
  surface->media_nb_height_tiles = height / surface->max_texture_size;
  height_upper_pot = get_upper_power_of_two (height);
  size = MIN (surface->max_texture_size, height_upper_pot);
  surface->media_height_texel_size = 1.0f / size;
  height_modulo = height % surface->max_texture_size;
  if (height_modulo)
    {
      surface->media_nb_height_tiles += 1;
      surface->media_last_tile_t = height_modulo / size;
    }

/*   g_print ("Number of tiles: %d, %d\n" */
/*            "Last tile texture coord: %.2f, %.2f\n", */
/*            surface->media_nb_width_tiles, surface->media_nb_height_tiles, */
/*            last_tile_s, last_tile_t); */

  /* The total number of tile */
  surface->media_nb_tiles = surface->media_nb_width_tiles
    * surface->media_nb_height_tiles;

  /* Texture object allocation */
  surface->media_texture =
    (PgmRenderTextureObject*) g_malloc0 (sizeof (PgmRenderTextureObject));
  if (format == PGM_RENDER_MEDIA_YV12 || format == PGM_RENDER_MEDIA_I420)
    {
      surface->media_texture->size = 3 * surface->media_nb_tiles;
      surface->media_texture->identifiants =
        (guint*) g_malloc0 (3 * surface->media_nb_tiles * sizeof (guint));
    }
  else
    {
      surface->media_texture->size = surface->media_nb_tiles;
      surface->media_texture->identifiants =
        (guint*) g_malloc0 (surface->media_nb_tiles * sizeof (guint));
    }

  /* Vertices allocation.
   * 12 is the size of (x,y,z) * (v1,v2,v3,v4) */
  surface->media_tile_vertex =
    (gfloat*) g_malloc0 (surface->media_nb_tiles * 12 * sizeof (gfloat));

  /* Texture coordinates allocation.
   * 8 is the size of (s,t) * (tc1,tc2,tc3,tc4) */
  surface->media_tile_tex_coord =
    (gfloat*) g_malloc (surface->media_nb_tiles * 8 * sizeof (gfloat));

  /* Fill media properties */
  surface->media_width = width;
  surface->media_pot_width = width_upper_pot;
  surface->media_height = height;
  surface->media_pot_height = height_upper_pot;
  surface->media_stride = stride;
  surface->media_master = surface;
  surface->media_format = format;

  if (surface->media_interp == PGM_RENDER_MEDIA_BILINEAR)
    set_texture_coordinates (surface, surface->media_width_texel_size,
                             surface->media_height_texel_size);
  else
    set_texture_coordinates (surface, 0.0f, 0.0f);
  set_pixel_aspect_ratio (surface);
}

static void
_set_image (PgmRenderMediaIFace *obj, PgmRenderMediaPixelFormat format,
            gint width, gint height, gint stride)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderGl1SurfaceClass *klass = PGM_RENDER_GL1_SURFACE_GET_CLASS (surface);

  gl1_surface_imedia_del_image (obj);

  if (surface->media_master->media_loaded)
    {
      if (surface->media_master == surface)
        gl1_surface_imedia_del_image (obj);

      /* It's a slave, let's clear the data and generate the geometry */
      if (surface->media_master != surface)
        {
          g_free (surface->media_tile_vertex);
          init_image_data (surface);
          generate_geometry_tiling (surface, format, width, height, stride);
        }
      /* Width and height are the same, we don't need to regenerate the
       * tiling, just to put the old texture identifiants in the trash */
      else if (surface->media_width == width
               && surface->media_height == height)
        {
          klass->texture_trash = g_list_prepend (klass->texture_trash,
                                                 surface->media_texture);

          surface->media_texture = (PgmRenderTextureObject*)
            g_malloc0 (sizeof (PgmRenderTextureObject));

          surface->media_texture->identifiants =
            (guint*) g_malloc0 (surface->media_nb_tiles * sizeof (guint));

          surface->media_texture->size = surface->media_nb_tiles;
        }
      /* It's a different size, we have to regenerate the tiling */
      else
        {
          klass->texture_trash = g_list_prepend (klass->texture_trash,
                                                 surface->media_texture);
          surface->media_texture = NULL;
          g_free (surface->media_tile_vertex);
          surface->media_tile_vertex = NULL;
          g_free (surface->media_tile_tex_coord);
          surface->media_tile_tex_coord = NULL;
          generate_geometry_tiling (surface, format, width, height, stride);
        }
    }
  else
    generate_geometry_tiling (surface, format, width, height, stride);
}

/* FIXME: If the surface is a master, the slaves must be changed accordingly */
static PgmRenderError
gl1_surface_imedia_set_image (PgmRenderMediaIFace *obj,
                              PgmRenderMediaPixelFormat format,
                              gint width, gint height, gint stride,
                              gconstpointer data)
{
  PgmRenderDrawable *drawable = PGM_RENDER_DRAWABLE (obj);
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderGl1SurfaceClass *klass = PGM_RENDER_GL1_SURFACE_GET_CLASS (surface);

  g_return_val_if_fail (klass->media_format_lut[format] != -1,
			PGM_RENDER_ERROR_X);

  _set_image (obj, format, width, height, stride);

  /* Copy and store the image data */
  PGM_RENDER_DRAWABLE_LOCK (surface);
  if (surface->media_data)
    g_free (surface->media_data);
  surface->media_data = g_memdup (data, width * height *
                                  klass->media_format_lut_size[format]);
  klass->surface_with_media_data = g_list_prepend
    (klass->surface_with_media_data, drawable);
  PGM_RENDER_DRAWABLE_UNLOCK (surface);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_set_gst_image (PgmRenderMediaIFace *obj,
				  PgmRenderMediaPixelFormat format, gint width,
				  gint height, gint stride, GstBuffer *buffer)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderGl1SurfaceClass *klass = PGM_RENDER_GL1_SURFACE_GET_CLASS (surface);
  PgmRenderGl1Surface *slave;
  GList *element;
  GstBuffer *tmp;

  g_return_val_if_fail (obj, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (buffer, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (width >= 0, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height >= 0, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (klass->media_format_lut[format] != -1,
			PGM_RENDER_ERROR_X);

  /* The stream is already loaded */
  if (surface->media_master->media_gst_loaded)
    {
      /* If the frame size is different, we regenerate the datas */
      if (width != surface->media_master->media_width
          || height != surface->media_master->media_height)
        {
          gl1_surface_imedia_del_image (obj);
          generate_geometry_tiling (surface, format, width, height, stride);
        }
    }
  else
    {
      /* Generate the datas */
      generate_geometry_tiling (surface, format, width, height, stride);

      /* Update the slave medias if it's a master */
      if (surface->media_master == surface)
        {
          element = surface->media_master->media_slave;
          while (element)
            {
              generate_geometry_tiling (element->data, surface->media_format,
                                        surface->media_width,
                                        surface->media_height,
                                        surface->media_stride);
              slave = PGM_RENDER_GL1_SURFACE (element->data);
              slave->media_gst_loaded = TRUE;
              slave->media_texture = surface->media_texture;
              element = g_list_next (element);
            }
        }
    }

  /* Store the GStreamer buffer */
  PGM_RENDER_DRAWABLE_LOCK (surface);
  gst_buffer_ref (buffer);
  tmp = surface->media_gst_buffer;
  surface->media_gst_buffer = buffer;
  if (tmp)
    gst_buffer_unref (tmp);
  PGM_RENDER_DRAWABLE_UNLOCK (surface);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_set_image_from_drawable (PgmRenderMediaIFace *obj,
					    PgmRenderDrawable *drawable)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderGl1Surface *master;

  g_return_val_if_fail (drawable, PGM_RENDER_ERROR_X);

  master = PGM_RENDER_GL1_SURFACE (drawable)->media_master;

  if (master)
    {
      gl1_surface_imedia_del_image (obj);
      surface->media_width = master->media_width;
      surface->media_pot_width = master->media_pot_width;
      surface->media_height = master->media_height;
      surface->media_pot_height = master->media_pot_height;
      surface->media_ratio = master->media_ratio;
      surface->media_loaded = master->media_loaded;
      surface->media_gst_loaded = master->media_gst_loaded;
      surface->media_format = master->media_format;
      surface->media_master = master;
      master->media_slave = g_list_prepend (master->media_slave, surface);
      surface->media_texture = master->media_texture;
      surface->media_nb_width_tiles = master->media_nb_width_tiles;
      surface->media_nb_height_tiles = master->media_nb_height_tiles;
      surface->media_nb_tiles = master->media_nb_tiles;
      surface->media_tile_vertex =
        (gfloat*) g_malloc0 (surface->media_nb_tiles * 12 * sizeof (gfloat));
      surface->media_tile_tex_coord = master->media_tile_tex_coord;
      set_pixel_aspect_ratio (surface);
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_del_image (PgmRenderMediaIFace *obj)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderGl1SurfaceClass *klass = PGM_RENDER_GL1_SURFACE_GET_CLASS (surface);

  GList *element = surface->media_slave;

  if (surface->media_texture)
    {
      /* Is it a master surface ? */
      if (surface->media_master == surface)
	{
          /* Store the TextureObject in the texture trash */
          klass->texture_trash = g_list_prepend (klass->texture_trash,
                                                 surface->media_texture);

          /* Reinit the slave surface(s) */
	  while (element)
	    {
              g_free (PGM_RENDER_GL1_SURFACE
                      (element->data)->media_tile_vertex);
	      init_image_data (element->data);
	      element = g_list_next (element);
	    }
          g_list_free (surface->media_slave);
          surface->media_slave = NULL;
	}

      g_free (surface->media_tile_tex_coord);
      g_free (surface->media_tile_vertex);

      /* Reinitialize the surface */
      init_image_data (surface);
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_set_interp_type (PgmRenderMediaIFace *obj,
				    PgmRenderMediaInterpType type)
{
  PgmRenderGl1Surface *surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderDrawable *drawable = PGM_RENDER_DRAWABLE (surface);
  PgmRenderGl1SurfaceClass *klass;

  if (surface->media_interp == type)
    return PGM_RENDER_ERROR_X;

  surface->media_interp = type;

  /* The I420 and YV12 cases need one pixel offset at the beginning and
   * the end of the tex coords to avoid showing green borders when the
   * filtering is set to BILINEAR. */
  if (surface->media_format == PGM_RENDER_MEDIA_I420
      || surface->media_format == PGM_RENDER_MEDIA_YV12)
    {
      if (type == PGM_RENDER_MEDIA_BILINEAR)
        set_texture_coordinates (surface, surface->media_width_texel_size,
                                 surface->media_height_texel_size);
      else
        set_texture_coordinates (surface, 0.0f, 0.0f);
    }

  if (surface->media_format != PGM_RENDER_MEDIA_YUYV
      && surface->media_format != PGM_RENDER_MEDIA_UYVY)
    {
      klass =  PGM_RENDER_GL1_SURFACE_GET_CLASS (surface);
      klass->surface_with_new_media_interp =
        g_list_prepend (klass->surface_with_new_media_interp, drawable);
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_imedia_get_interp_type (PgmRenderMediaIFace *obj,
				    PgmRenderMediaInterpType *type)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (type, PGM_RENDER_ERROR_X);

  *type = gl1_surface->media_interp;

  return PGM_RENDER_ERROR_X;
}

static void
media_iface_init (PgmRenderMediaIFaceClass *iface)
{
  iface->set_color = gl1_surface_imedia_set_color;
  iface->get_color = gl1_surface_imedia_get_color;
  iface->set_image = gl1_surface_imedia_set_image;
  iface->set_gst_image = gl1_surface_imedia_set_gst_image;
  iface->set_image_from_drawable = gl1_surface_imedia_set_image_from_drawable;
  iface->del_image = gl1_surface_imedia_del_image;
  iface->set_interp_type = gl1_surface_imedia_set_interp_type;
  iface->get_interp_type = gl1_surface_imedia_get_interp_type;
}

/* Shape interface implementation */

static PgmRenderError
gl1_surface_ishape_set_arc (PgmRenderShapeIFace *obj, gfloat width,
			    gfloat height)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);
  gfloat value = G_PI_2;
  gfloat offset = G_PI / 19.5f;
  gfloat mid_size;
  gint i;

  mid_size = (gl1_surface->vertex[3].x - gl1_surface->vertex[0].x) * 0.5f;
  gl1_surface->arc_width = MIN (width, mid_size);
  mid_size = (gl1_surface->vertex[1].y - gl1_surface->vertex[0].y) * 0.5f;
  gl1_surface->arc_height = MIN (height, mid_size);

  gl1_surface->use_arc = !((width == 0.0f) && (height == 0.0f));

  if (gl1_surface->use_arc)
    {
      for (i = 0; i < 40; i++)
	{
	  gl1_surface->arc_vertex[i].x = cos (value) * gl1_surface->arc_width;
	  gl1_surface->arc_vertex[i].y = sin (value) * gl1_surface->arc_height;
	  value += offset;
	}
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ishape_get_arc (PgmRenderShapeIFace *obj, gfloat *width,
			    gfloat *height)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (height, PGM_RENDER_ERROR_X);

  *width = gl1_surface->arc_width;
  *height = gl1_surface->arc_height;

  return PGM_RENDER_ERROR_X;
}

static void
shape_iface_init (PgmRenderShapeIFaceClass *iface)
{
  iface->set_arc = gl1_surface_ishape_set_arc;
  iface->get_arc = gl1_surface_ishape_get_arc;
}

/* Stroke interface implementation */

static PgmRenderError
gl1_surface_istroke_set_width (PgmRenderStrokeIFace *obj, gfloat width)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  /* FIXME: Use an opcode queue of GL commands */
  gl1_surface->stroke_width = MAX (0.0f, width);
/*   gl->line_width (gl1_surface->stroke_width); */

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_istroke_get_width (PgmRenderStrokeIFace *obj, gfloat *width)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (width, PGM_RENDER_ERROR_X);

  *width = gl1_surface->stroke_width;

  return PGM_RENDER_ERROR_X;
}

/* FIXME: Doesn't work */
static PgmRenderError
gl1_surface_istroke_set_pattern (PgmRenderStrokeIFace *obj,
				 PgmRenderStrokePattern pattern)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->stroke_pattern = pattern;

  switch (pattern)
    {
    case PGM_RENDER_STROKE_CONTINUOUS:
      /* glDisable (GL_LINE_STIPPLE); */
      break;

    case PGM_RENDER_STROKE_DASHED:
      /* glEnable (GL_LINE_STIPPLE); */
      /* glLineStipple (0x00FF, gl1_surface->stroke_factor); */
      break;

    default:
      break;
    }

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_istroke_get_pattern (PgmRenderStrokeIFace *obj,
				 PgmRenderStrokePattern *pattern)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (pattern, PGM_RENDER_ERROR_X);

  *pattern = gl1_surface->stroke_pattern;

  return PGM_RENDER_ERROR_X;
}

/* FIXME: Doesn't work */
static PgmRenderError
gl1_surface_istroke_set_factor (PgmRenderStrokeIFace *obj, gint factor)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->stroke_factor = factor;
  /* glLineStipple (0x00FF, factor); */

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_istroke_get_factor (PgmRenderStrokeIFace *obj, gint *factor)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (factor, PGM_RENDER_ERROR_X);

  *factor = gl1_surface->stroke_factor;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_istroke_set_color (PgmRenderStrokeIFace *obj, guchar r,
			       guchar g, guchar b, guchar a)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->stroke_color.r = (gfloat) r * INV_255;
  gl1_surface->stroke_color.g = (gfloat) g * INV_255;
  gl1_surface->stroke_color.b = (gfloat) b * INV_255;
  gl1_surface->stroke_color.a = (gfloat) a * INV_255;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_istroke_get_color (PgmRenderStrokeIFace *obj, guchar *r,
			       guchar *g, guchar *b, guchar *a)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (r, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (g, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (b, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (a, PGM_RENDER_ERROR_X);

  *r = (guchar) (gl1_surface->stroke_color.r * 255);
  *g = (guchar) (gl1_surface->stroke_color.g * 255);
  *b = (guchar) (gl1_surface->stroke_color.b * 255);
  *a = (guchar) (gl1_surface->stroke_color.a * 255);

  return PGM_RENDER_ERROR_X;
}

static void
stroke_iface_init (PgmRenderStrokeIFaceClass *iface)
{
  iface->set_width = gl1_surface_istroke_set_width;
  iface->get_width = gl1_surface_istroke_get_width;
  iface->set_pattern = gl1_surface_istroke_set_pattern;
  iface->get_pattern = gl1_surface_istroke_get_pattern;
  iface->set_factor = gl1_surface_istroke_set_factor;
  iface->get_factor = gl1_surface_istroke_get_factor;
  iface->set_color = gl1_surface_istroke_set_color;
  iface->get_color = gl1_surface_istroke_get_color;
}

/* Pipe3d interface implementation */

static PgmRenderError
gl1_surface_ipipe3d_set_translation (PgmRenderPipe3dIFace *obj, gfloat x,
				     gfloat y, gfloat z)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);
  PgmRenderDrawable *drawable = PGM_RENDER_DRAWABLE (obj);

  gl1_surface->translation[0] = x;
  gl1_surface->translation[1] = y;
  gl1_surface->translation[2] = z;

  if (drawable->window && drawable->order == PGM_RENDER_DRAWABLE_Z_ORDERED)
    pgm_render_base_window_reorder_drawable (drawable->window, drawable, 0);

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ipipe3d_get_translation (PgmRenderPipe3dIFace *obj, gfloat *x,
				     gfloat *y, gfloat *z)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (z, PGM_RENDER_ERROR_X);

  *x = gl1_surface->translation[0];
  *y = gl1_surface->translation[1];
  *z = gl1_surface->translation[2];

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ipipe3d_set_rotation (PgmRenderPipe3dIFace *obj, gfloat x,
				  gfloat y, gfloat z)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  gl1_surface->rotation[0] = x;
  gl1_surface->rotation[1] = y;
  gl1_surface->rotation[2] = z;

  return PGM_RENDER_ERROR_X;
}

static PgmRenderError
gl1_surface_ipipe3d_get_rotation (PgmRenderPipe3dIFace *obj, gfloat *x,
				  gfloat *y, gfloat *z)
{
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (obj);

  g_return_val_if_fail (x, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (y, PGM_RENDER_ERROR_X);
  g_return_val_if_fail (z, PGM_RENDER_ERROR_X);

  *x = gl1_surface->rotation[0];
  *y = gl1_surface->rotation[1];
  *z = gl1_surface->rotation[2];

  return PGM_RENDER_ERROR_X;
}

static void
pipe3d_iface_init (PgmRenderPipe3dIFaceClass *iface)
{
  iface->set_translation = gl1_surface_ipipe3d_set_translation;
  iface->get_translation = gl1_surface_ipipe3d_get_translation;
  iface->set_rotation = gl1_surface_ipipe3d_set_rotation;
  iface->get_rotation = gl1_surface_ipipe3d_get_rotation;
}

/* Drawable base class method overloading */

/* FIXME: Use vertex arrays and VBOs */
static PgmRenderError
pgm_render_gl1_surface_draw (PgmRenderDrawable *drawable)
{
  PgmRenderGl1Window *window = PGM_RENDER_GL1_WINDOW (drawable->window);
  PgmRenderGlProcAddressList *gl = window->backend->proc_address;
  PgmRenderGl1Surface *gl1_surface = PGM_RENDER_GL1_SURFACE (drawable);
  PgmRenderDrawableClass *klass = PGM_RENDER_DRAWABLE_GET_CLASS (drawable);
  static gboolean caps_set = FALSE;
  ColorSpaceShader shader;
  gfloat x_offset, y_offset;
  gint i;

  /* Add the YUV caps if the shaders are supported. */
  /* FIXME: Don't do that here */
  if (G_UNLIKELY (!caps_set))
    {
      if (window->backend->feature_mask & PGM_RENDER_GL_FEATURE_GLSL)
        klass->caps = gst_caps_from_string (PGM_RENDER_GL1_CAPS_RGB " ; "
                                            PGM_RENDER_GL1_CAPS_YUV);
      caps_set = TRUE;
    }

  PGM_RENDER_DRAWABLE_LOCK (drawable);
  /* Compute the offset to the center of the polygon. */
  /* FIXME: Should be precomputed. */
  /* FIXME: Access to tiles should use macros. */
  if (gl1_surface->media_master->media_loaded
      || gl1_surface->media_master->media_gst_loaded)
    {
      x_offset = (gl1_surface->media_tile_vertex
                  [(gl1_surface->media_nb_height_tiles) * 12 *
                   (gl1_surface->media_nb_width_tiles-1) + 9]
                  + gl1_surface->media_tile_vertex[0]) * 0.5f;
      y_offset = (gl1_surface->media_tile_vertex
                  [(gl1_surface->media_nb_height_tiles-1) * 12 + 4]
                  + gl1_surface->media_tile_vertex[1]) * 0.5f;
    }
  else
    {
      x_offset = (gl1_surface->vertex[3].x + gl1_surface->vertex[0].x) * 0.5f;
      y_offset = (gl1_surface->vertex[1].y + gl1_surface->vertex[0].y) * 0.5f;
    }

  gl->load_identity ();
  gl->translate_f (gl1_surface->translation[0] + x_offset,
                   gl1_surface->translation[1] + y_offset,
                   gl1_surface->translation[2]);
  gl->rotate_f (gl1_surface->rotation[0], 1.0f, 0.0f, 0.0f);
  gl->rotate_f (gl1_surface->rotation[1], 0.0f, 1.0f, 0.0f);
  gl->rotate_f (gl1_surface->rotation[2], 0.0f, 0.0f, 1.0f);
  gl->translate_f (-x_offset, -y_offset, 0.0f);

  /* Draw the texture if a media is loaded */
  if (gl1_surface->media_master->media_loaded
      || gl1_surface->media_master->media_gst_loaded)
    {
      /* FIXME: It's a CPU angry quick fix to set the correct p-a-r to
       *        the new levels of Elisa. We need to find a solution ASAP. */
      set_pixel_aspect_ratio (gl1_surface);

      gl->color_4f (gl1_surface->color[0].r, gl1_surface->color[0].g,
                    gl1_surface->color[0].b, gl1_surface->color[0].a);
      gl->enable (PGM_RENDER_GL_TEXTURE_2D);
      gl->enable_client_state (PGM_RENDER_GL_VERTEX_ARRAY);
      gl->enable_client_state (PGM_RENDER_GL_TEXTURE_COORD_ARRAY);
      gl->vertex_pointer (3, PGM_RENDER_GL_FLOAT, 0,
                          gl1_surface->media_tile_vertex);
      gl->tex_coord_pointer (2, PGM_RENDER_GL_FLOAT, 0,
                             gl1_surface->media_master->media_tile_tex_coord);

      /* There's support for GLSL shaders */
      if (window->backend->feature_mask & PGM_RENDER_GL_FEATURE_GLSL)
        {
          /* I420 or YV12 color spaces using multitexturing */
          if (gl1_surface->media_format == PGM_RENDER_MEDIA_I420
              || gl1_surface->media_format == PGM_RENDER_MEDIA_YV12)
            {
              shader = window->color_space_shader
                [gl1_surface->media_format][0/* gl1_surface->media_interp */];
              gl->use_program_object (shader.program);
              shader.tex_width_loc = gl->get_uniform_location (shader.program,
                                                               "tex_width");
              gl->uniform_1f (shader.tex_width_loc, gl1_surface->media_pot_width);
              shader.tex_height_loc = gl->get_uniform_location (shader.program,
                                                                "tex_height");
              gl->uniform_1f (shader.tex_height_loc,
                              gl1_surface->media_pot_height);
              shader.tex_loc[0] = gl->get_uniform_location (shader.program,
                                                            "tex0");
              gl->uniform_1i (shader.tex_loc[0], 0);
              shader.tex_loc[1] = gl->get_uniform_location (shader.program,
                                                            "tex1");
              gl->uniform_1i (shader.tex_loc[1], 1);
              shader.tex_loc[2] = gl->get_uniform_location (shader.program,
                                                            "tex2");
              gl->uniform_1i (shader.tex_loc[2], 2);
              for (i = 0; i < gl1_surface->media_master->media_nb_tiles; i++)
                {
                  gl->active_texture (PGM_RENDER_GL_TEXTURE0);
                  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                    gl1_surface->media_master->media_texture->
                                    identifiants[3*i]);
                  gl->active_texture (PGM_RENDER_GL_TEXTURE1);
                  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                    gl1_surface->media_master->media_texture->
                                    identifiants[3*i+1]);
                  gl->active_texture (PGM_RENDER_GL_TEXTURE2);
                  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                    gl1_surface->media_master->media_texture->
                                    identifiants[3*i+2]);
                  gl->draw_arrays (PGM_RENDER_GL_QUADS, i*4, 4);
                }
              gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, 0);
              gl->active_texture (PGM_RENDER_GL_TEXTURE1);
              gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, 0);
              gl->active_texture (PGM_RENDER_GL_TEXTURE0);
              gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D, 0);
              gl->use_program_object (0);
            }

          /* YUYV or UYVY */
          else if (gl1_surface->media_format == PGM_RENDER_MEDIA_UYVY
                   || gl1_surface->media_format == PGM_RENDER_MEDIA_YUYV)
            {
              shader = window->color_space_shader
                [gl1_surface->media_format][0/* gl1_surface->media_interp */];
              gl->use_program_object (shader.program);
              shader.tex_width_loc = gl->get_uniform_location (shader.program,
                                                               "tex_width");
              gl->uniform_1f (shader.tex_width_loc,
                              gl1_surface->media_pot_width * 0.5f);
              shader.tex_height_loc = gl->get_uniform_location (shader.program,
                                                                "tex_height");
              gl->uniform_1f (shader.tex_height_loc,
                              gl1_surface->media_pot_height);
              for (i = 0; i < gl1_surface->media_master->media_nb_tiles; i++)
                {
                  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                    gl1_surface->media_master->media_texture->
                                    identifiants[i]);
                  gl->draw_arrays (PGM_RENDER_GL_QUADS, i*4, 4);
                }
              gl->use_program_object (0);
            }

          /* RGB */
          else
            {
              shader = window->color_space_shader
                [gl1_surface->media_format][0/* gl1_surface->media_interp */];
              gl->use_program_object (shader.program);
              shader.tex_width_loc = gl->get_uniform_location (shader.program,
                                                               "tex_width");
              gl->uniform_1f (shader.tex_width_loc,
                              gl1_surface->media_pot_width);
              shader.tex_height_loc = gl->get_uniform_location (shader.program,
                                                                "tex_height");
              gl->uniform_1f (shader.tex_height_loc,
                              gl1_surface->media_pot_height);
              for (i = 0; i < gl1_surface->media_master->media_nb_tiles; i++)
                {
                  gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                    gl1_surface->media_master->media_texture->
                                    identifiants[i]);
                  gl->draw_arrays (PGM_RENDER_GL_QUADS, i*4, 4);
                }
              gl->use_program_object (0);
            }
        }

      /* There's no support for GLSL shaders */
      else
        {
          for (i = 0; i < gl1_surface->media_master->media_nb_tiles; i++)
            {
              gl->bind_texture (PGM_RENDER_GL_TEXTURE_2D,
                                gl1_surface->media_master->media_texture->
                                identifiants[i]);
              gl->draw_arrays (PGM_RENDER_GL_QUADS, i*4, 4);
            }
        }

      gl->disable_client_state (PGM_RENDER_GL_VERTEX_ARRAY);
      gl->disable_client_state (PGM_RENDER_GL_TEXTURE_COORD_ARRAY);
      gl->disable (PGM_RENDER_GL_TEXTURE_2D);
    }

  /* Draw the shape */
  else
    {
      if (!gl1_surface->use_arc)
	{
	  gl->begin (PGM_RENDER_GL_QUADS);
	  for (i = 0; i < 4; i++)
	    {
	      gl->color_4f (gl1_surface->color[i].r,
                            gl1_surface->color[i].g,
                            gl1_surface->color[i].b,
                            gl1_surface->color[i].a);
	      gl->vertex_3f (gl1_surface->vertex[i].x,
                             gl1_surface->vertex[i].y,
                             gl1_surface->vertex[i].z);
	    }
	  gl->end ();
	  if (gl1_surface->stroke_width > 0.0f)
	    {
	      gl->color_4f (gl1_surface->stroke_color.r,
                            gl1_surface->stroke_color.g,
                            gl1_surface->stroke_color.b,
                            gl1_surface->stroke_color.a);
              gl->enable (PGM_RENDER_GL_LINE_SMOOTH);
	      gl->disable (PGM_RENDER_GL_DEPTH_TEST);
	      gl->begin (PGM_RENDER_GL_LINE_LOOP);
	      for (i = 0; i < 4; i++)
		{
		  gl->vertex_3f (gl1_surface->vertex[i].x,
                                 gl1_surface->vertex[i].y,
                                 gl1_surface->vertex[i].z);
		}
	      gl->end ();
              gl->disable (PGM_RENDER_GL_LINE_SMOOTH);
	      gl->enable (PGM_RENDER_GL_DEPTH_TEST);
	    }
	}
      else
	{
	  gl->begin (PGM_RENDER_GL_POLYGON);
	  gl->color_4f (gl1_surface->color[0].r,
                        gl1_surface->color[0].g,
                        gl1_surface->color[0].b,
                        gl1_surface->color[0].a);
	  for (i = 0; i < 10; i++)
	    {
	      gl->vertex_3f (gl1_surface->vertex[0].x
                             + gl1_surface->arc_vertex[i].x
                             + gl1_surface->arc_width,
                             (gl1_surface->vertex[0].y
                              - gl1_surface->arc_vertex[i].y
                              + gl1_surface->arc_height),
                             gl1_surface->vertex[0].z);
	    }
	  gl->color_4f (gl1_surface->color[1].r,
                        gl1_surface->color[1].g,
                        gl1_surface->color[1].b,
                        gl1_surface->color[1].a);
	  for (i = 10; i < 20; i++)
	    {
	      gl->vertex_3f (gl1_surface->vertex[1].x
                             + gl1_surface->arc_vertex[i].x
                             + gl1_surface->arc_width,
                             (gl1_surface->vertex[1].y
                              - gl1_surface->arc_vertex[i].y
                              - gl1_surface->arc_height),
                             gl1_surface->vertex[1].z);
	    }
	  gl->color_4f (gl1_surface->color[2].r,
                        gl1_surface->color[2].g,
                        gl1_surface->color[2].b,
                        gl1_surface->color[2].a);
	  for (i = 20; i < 30; i++)
	    {
	      gl->vertex_3f (gl1_surface->vertex[2].x
                             + gl1_surface->arc_vertex[i].x
                             - gl1_surface->arc_width,
                             (gl1_surface->vertex[2].y
                              - gl1_surface->arc_vertex[i].y
                              - gl1_surface->arc_height),
                             gl1_surface->vertex[2].z);
	    }
	  gl->color_4f (gl1_surface->color[3].r,
                        gl1_surface->color[3].g,
                        gl1_surface->color[3].b,
                        gl1_surface->color[3].a);
	  for (i = 30; i < 40; i++)
	    {
	      gl->vertex_3f (gl1_surface->vertex[3].x
                             + gl1_surface->arc_vertex[i].x
                             - gl1_surface->arc_width,
                             (gl1_surface->vertex[3].y
                              - gl1_surface->arc_vertex[i].y
                              + gl1_surface->arc_height),
                             gl1_surface->vertex[3].z);
	    }
	  gl->end ();
	  
	  if (gl1_surface->stroke_width > 0.0f)
	    {
	      gl->disable (PGM_RENDER_GL_DEPTH_TEST);
              gl->enable (PGM_RENDER_GL_LINE_SMOOTH);
	      gl->line_width (gl1_surface->stroke_width);
	      gl->color_4f (gl1_surface->stroke_color.r,
                            gl1_surface->stroke_color.g,
                            gl1_surface->stroke_color.b,
                            gl1_surface->stroke_color.a);
	      gl->begin (PGM_RENDER_GL_LINE_LOOP);
	      for (i = 0; i < 10; i++)
		{
		  gl->vertex_3f (gl1_surface->vertex[0].x
                                 + gl1_surface->arc_vertex[i].x
                                 + gl1_surface->arc_width,
                                 (gl1_surface->vertex[0].y
                                  - gl1_surface->arc_vertex[i].y
                                  + gl1_surface->arc_height),
                                 gl1_surface->vertex[0].z);
		}
	      for (i = 10; i < 20; i++)
		{
		  gl->vertex_3f (gl1_surface->vertex[1].x
                                 + gl1_surface->arc_vertex[i].x
                                 + gl1_surface->arc_width,
                                 (gl1_surface->vertex[1].y
                                  - gl1_surface->arc_vertex[i].y
                                  - gl1_surface->arc_height),
                                 gl1_surface->vertex[1].z);
		}
	      for (i = 20; i < 30; i++)
		{
		  gl->vertex_3f (gl1_surface->vertex[2].x
                                 + gl1_surface->arc_vertex[i].x
                                 - gl1_surface->arc_width,
                                 (gl1_surface->vertex[2].y
                                  - gl1_surface->arc_vertex[i].y
                                  - gl1_surface->arc_height),
                                 gl1_surface->vertex[2].z);
		}
	      for (i = 30; i < 40; i++)
		{
		  gl->vertex_3f (gl1_surface->vertex[3].x
                                 + gl1_surface->arc_vertex[i].x
                                 - gl1_surface->arc_width,
                                 (gl1_surface->vertex[3].y
                                  - gl1_surface->arc_vertex[i].y
                                  + gl1_surface->arc_height),
                                 gl1_surface->vertex[3].z);
		}
	      gl->end ();
              gl->disable (PGM_RENDER_GL_LINE_SMOOTH);
	      gl->enable (PGM_RENDER_GL_DEPTH_TEST);
	    }
	}
    }
  PGM_RENDER_DRAWABLE_UNLOCK (drawable);

  return PGM_RENDER_ERROR_X;
}

void
pgm_render_gl1_surface_set_scale (PgmRenderGl1Surface *surface, gfloat sx,
                                  gfloat sy)
{
  surface->sx = sx;
  surface->sy = sy;

  surface->vertex[0].x = surface->vertex[1].x =
    surface->layout_x_position * sx;
  surface->vertex[2].x = surface->vertex[3].x =
    surface->vertex[0].x + surface->layout_width * sx;
  surface->vertex[0].y = surface->vertex[3].y =
    surface->layout_y_position * sy;
  surface->vertex[1].y = surface->vertex[2].y =
    surface->vertex[0].y + surface->layout_height * sy;

  if (surface->media_loaded || surface->media_gst_loaded)
    set_pixel_aspect_ratio (surface);
}
